/*******************************************************************************
 * Copyright (c) 2003, 2007 s IT Solutions AT Spardat GmbH .
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/

/*
 * @(#) $Id: TableLayoutManager.java 5024 2010-02-10 16:19:12Z gub $
 *
 *
 *
 *
 *
 */
package at.spardat.xma.mdl.table;

import org.eclipse.swt.events.ControlEvent;
import org.eclipse.swt.events.ControlListener;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;

/**
 * TableLayoutManager implements a more sophisticated Layout for Table columns than just column widths. <br/>
 * For each column the follwing can be defined:
 * <ul>
 * <li> The absolut width in pixel.
 * <li> The width as percentage of the width of the whole table.
 * <li> The minimum width in pixel.
 * <li> The maximum width in pixel.
 * </ul>
 * Only the absolut width or a percentage may be specified for each individual column.<br>
 * Everytime, the table is resized, this information is used to calculate new widths of the columns.
 * If the end user resizes a comlumn header via the GUI, the absolute width or percentage is
 * ajusted automatically. If the end user violates the minimum or maximum, the respective limit is
 * ajusted, too. Only the minLimit is enforced to prevent the user from completly removing a column.
 * <p>
 * A default width for columns without absolut width or percentage definition may be defined.
 * Default minimum and maximun width for columns without individal limits may be defined, too.
 * <br>
 * One column may be specified to strech over the rest of the table width. Its absolute width
 * or percentage is ignored. Only its limits are respected.
 */
public class TableLayoutManager implements ControlListener {
    /**
     * Use no information from the table.
     */
    public static final int NONE    = 0;
    /**
     * Calculate the percentage of the column based on its current width and the current width of the table.
     */
    public static final int PERCENT = 1;
    /**
     * Determine that the calculation of percentages should always yield a sum of 100 percent.
     */
    public static final int SCALED = 1<<4;
    /**
     * Calculate the percentage of the column based on its current width and the sum of all column widths.
     * This always yields a sum of 100 percent. <br>
     * SCALEDPERCENT = PERCENT | SCALED
     */
    public static final int SCALEDPERCENT = PERCENT | SCALED;
    /**
     * Use the currend width of the table as it's minimum width.
     */
    public static final int MIN =  1<<1;
    /**
     * Use the currend width of the table as it's maximum width.
     */
    public static final int MAX =  1<<2;
    /**
     * Use the currend width of the table as it's absolut width.
     */
    public static final int ABSOLUT = 1<<3;

    int[] min,max,absolut;
    float[] percent;
    int minLimit=10;
    Table table;
    int tableWidth;
    int strech = -1;
    int strechWidth;
    int defaultMinWidth=minLimit;
    int defaultMaxWidth=Short.MAX_VALUE;
    int defaultAbsolutWidth=40;
    boolean layouting=false;

    /**
     * Attaches the TableLayoutManager to an SWT-Table. From this time on, the column widths are controled by
     * the TableLayoutManager.
     * @param table to manage
     * @param take defines which layout information should be calculated from the current width of the columns.
     *  possible values: {@link #PERCENT}, {@link #SCALEDPERCENT}, {@link #MIN}, {@link #MAX}, {@link #ABSOLUT}. These value may be ored
     *  together to take more information.
     */
    public void manageTable(Table table,int take) {
        if(this.table!=null) { // remove from old table if any
            this.table.removeControlListener(this);
            for(int i=0,size=this.table.getColumnCount();i<size;i++) {
                this.table.getColumn(i).removeControlListener(this);
            }
        }
        this.table=table;
        this.table.addControlListener(this);
        for(int i=0,size=this.table.getColumnCount();i<size;i++) {
            this.table.getColumn(i).addControlListener(this);
        }
        recalcLayoutInfoFromTable(take);
    }

    /**
     * Recalculates all layout informations from the current width of the columns of the attached table.
     * This method can be used to change from absolute widths to percentages or the other way round.
     * This method is used internally from {@link #manageTable(Table, int)}.
     * @param take defines which layout information should be calculated from the current width of the columns.
     *  possible values: {@link #PERCENT}, {@link #SCALEDPERCENT}, {@link #MIN}, {@link #MAX}, {@link #ABSOLUT}. These value may be ored
     *  together to take more information.
     */
    public void recalcLayoutInfoFromTable(int take) {
        int size=table.getColumnCount();
        min = new int[size];
        max = new int[size];
        absolut = new int[size];
        percent = new float[size];
        tableWidth = getTableClientWidth(table);
        int referenceWidth = tableWidth;
        if((take&SCALED)!=0) { // use sum of column widths instead
            referenceWidth = 0;
            for(int i=0;i<size;i++) {
                referenceWidth+=table.getColumn(i).getWidth();
            }
        }
        for(int i=0;i<size;i++) {
            TableColumn column = table.getColumn(i);
            if((take&PERCENT)!=0) percent[i]=(float)column.getWidth()/referenceWidth*100;
            else percent[i]=-1;
            if((take&MIN)!=0) {
                min[i]=column.getWidth();
                if(min[i]<minLimit) min[i]=minLimit;
            }
            else min[i] = defaultMinWidth;
            if((take&MAX)!=0) max[i]=column.getWidth();
            else max[i] = defaultMaxWidth;
            if((take&ABSOLUT)!=0) absolut[i]=column.getWidth();
            else if((take&PERCENT)!=0) absolut[i]=-1;
            else absolut[i] = defaultAbsolutWidth;
        }
    }

    /**
     * Adds a column to the TableLayoutManager. If you add a column to a managed table after calling
     * {@link #manageTable}, you have to call this method to inform the TableLayoutManager about the
     * new column.
     * @param column index of the new column in the SWT-Table.
     */
    public void addColumn(int column) {
        float[] newpercent = new float[percent.length+1];
        System.arraycopy(percent,0,newpercent,0,column);
        newpercent[column]=-1;
        System.arraycopy(percent,column,newpercent,column+1,percent.length-column);
        percent = newpercent;
        int[] newint = new int[min.length+1];
        System.arraycopy(min,0,newint,0,column);
        newint[column]=defaultMinWidth;
        System.arraycopy(min,column,newint,column+1,min.length-column);
        min = newint;
        newint = new int[max.length+1];
        System.arraycopy(max,0,newint,0,column);
        newint[column]=defaultMaxWidth;
        System.arraycopy(max,column,newint,column+1,max.length-column);
        max = newint;
        newint = new int[absolut.length+1];
        System.arraycopy(absolut,0,newint,0,column);
        newint[column]=defaultAbsolutWidth;
        System.arraycopy(absolut,column,newint,column+1,absolut.length-column);
        absolut = newint;
        table.getColumn(column).addControlListener(this);
    }

    /**
     * Removes a column to the TableLayoutManager. If you remove a column from a managed table after calling
     * {@link #manageTable}, you have to call this method to inform the TableLayoutManager about the
     * removal of the column.
     * @param column index of the column in the SWT-Table.
     */
    public void removeColumn(int column) {
        float[] newpercent = new float[percent.length-1];
        System.arraycopy(percent,0,newpercent,0,column);
        System.arraycopy(percent,column+1,newpercent,column,percent.length-column-1);
        percent = newpercent;
        int[] newint = new int[min.length-1];
        System.arraycopy(min,0,newint,0,column);
        System.arraycopy(min,column+1,newint,column,min.length-column-1);
        min = newint;
        newint = new int[max.length-1];
        System.arraycopy(max,0,newint,0,column);
        System.arraycopy(max,column+1,newint,column,max.length-column-1);
        max = newint;
        newint = new int[absolut.length-1];
        System.arraycopy(absolut,0,newint,0,column);
        System.arraycopy(absolut,column+1,newint,column,absolut.length-column-1);
        absolut = newint;
    }

    /**
     * Get the percentage of the given column. If no percentage is defined -1 is returned.
     * @param column index of the column in the SWT-Table.
     */
    public float getPercent(int column)  {
        return percent[column];
    }

    /**
     * Set the width of the given column in percent of the table width. The actual width of the
     * column is ajusted by the TableLayoutManager everytime the table is resized.
     * @param column index of the column in the SWT-Table.
     * @param percent column width/table width * 100 <br> -1 means remove the percentage information.
    */
    public void setPercent(int column,float percent) {
        this.percent[column]=percent;
        this.absolut[column]=-1;
    }

    /**
     * Get the minimum width of the given column in pixel.
     * @param column index of the column in the SWT-Table.
     */
    public int getMinWidth(int column) {
        return min[column];
    }

    /**
     * Set the minimum width of the given column in pixel.
     * @param column index of the column in the SWT-Table.
     */
    public void setMinWidth(int column, int min) {
//        if(min<minLimit) min=minLimit;
        this.min[column]=min;
    }

    /**
     * Get the maximum width of the given column in pixel.
     * @param column index of the column in the SWT-Table.
     */
    public int getMaxWidth(int column) {
        return max[column];
    }

    /**
     * Set the maximum width of the given column in pixel.
     * @param column index of the column in the SWT-Table.
     */
    public void setMaxWidth(int column, int max) {
        this.max[column]=max;
    }

    /**
     * Get the absolute width of the given column in pixel. If no absolute width is defined
     * in the TableLayoutManager, -1 is returned.
     * @param column index of the column in the SWT-Table.
     */
    public int getAbsolutWidth(int column) {
        return absolut[column];
    }

    /**
     * Set the absolute width of the given column in pixel. Setting a absolute width of a column,
     * tells the TableLayoutManager to leave the width of the given column unchanged, whenever the table
     * is resized.
     * @param column index of the column in the SWT-Table.
     * @param width fixed witdh of the given column. -1 removes the fixed width information.
     */
    public void setAbsolutWidth(int column, int width) {
        absolut[column]=width;
        percent[column]=-1;
    }

    /**
     * Set the column, which should take the whole space of the table which remains after calculating the
     * width of all other columns.
     * @param column index of the column to strech over the remaining space. -1 means strech no column.
     */
    public void setStrechColumn(int column) {
        strech = column;
    }

    /**
     * Scale the percentages of all columns by the given factor
     * @param factor to multiply to all percentages
     */
    public void scaleAllPercents(float factor) {
        for(int i=0;i<percent.length;i++) {
            if(percent[i]>=0) percent[i]*=factor;
        }
    }

    /**
     * Return the lower limit for all column widths. If the end user ties to resize any
     * column to a smaler width, this column width will be set to minLimit.
     */
    public int getMinLimit() {
        return minLimit;
    }

    /**
     * Set the lower limit for all column widths. If the end user ties to resize any
     * column to a smaler width, this column width will be set to minLimit.
     */
    public void setMinLimit(int minLimit) {
        this.minLimit = minLimit;
        if(defaultMinWidth<minLimit) defaultMinWidth=minLimit;
        for(int i=0;i<min.length;i++) {
            if(min[i]<minLimit) min[i]=minLimit;
        }
    }

    /**
     * Get the default absolute width.
     */
    public int getDefaultAbsolutWidth() {
        return defaultAbsolutWidth;
    }

    /**
     * Set the default absolute width. If set, all columns without absolut width or percentage information
     * get this width. They are not resized if the table is resized.
     */
    public void setDefaultAbsolutWidth(int defaultAbsolutWidth) {
        this.defaultAbsolutWidth = defaultAbsolutWidth;
    }

    /**
     * Get the default maximum width.
     */
    public int getDefaultMaxWidth() {
        return defaultMaxWidth;
    }

    /**
     * Set the default maximum width, used for all columns without individual maximum width.
     */
    public void setDefaultMaxWidth(int defaultMaxWidth) {
        this.defaultMaxWidth = defaultMaxWidth;
    }

    /**
     * Get the default minimum width.
     */
    public int getDefaultMinWidth() {
        return defaultMinWidth;
    }

    /**
     * Set the default minimum width, used for all columns without individual maximum width.
     */
    public void setDefaultMinWidth(int defaultMinWidth) {
        if(defaultMinWidth<minLimit) defaultMinWidth=minLimit;
        this.defaultMinWidth = defaultMinWidth;
    }

    /**
     * Get the width of the client area of the given table. This means the width of the table without
     * borders and scrollbars.
     */
    public static int getTableClientWidth(Table table) {
//        int width = table.getSize().x;
//        width -= table.getBorderWidth()*2;
////        //
////        Rectangle ca = table.getClientArea();
////        int count = table.getItemCount();
////        int itemHeigth = table.getItemHeight();
////        int inner = count*itemHeigth;
////        if(inner>ca.height) {
////        //
//            ScrollBar bar = table.getVerticalBar();
//            if(bar!=null) width -= bar.getSize().x;
////        }
//        return width;
        return table.getClientArea().width;
    }

    /**
     * Recalculate the width of all columns according to the layout information  of this TableLayoutManager.
     * The SWT-columns are resized to the new widths. Same as {@link #layout(boolean) layout(true)}.
     */
    public void layout() {
        layout(true);
    }

    /**
     * Recalculate the width of all columns according to the layout information  of this TableLayoutManager.
     * The SWT-columns are resized to the new widths.
     * @param force if false recalculate the column widths if the table width changed since the last calculation.
     *      if true recalculate in any case.
     */
    public void layout(boolean force) {
        int newTableWidth = getTableClientWidth(table);
        if(force||tableWidth!=newTableWidth) {
            layouting=true;
            if(percent.length!=table.getColumnCount()) throw new IllegalStateException();

            tableWidth=newTableWidth;
            int restWidth = tableWidth;
            for(int i=0;i<percent.length;i++) {
                if(i!=strech) {
                    int width=calcWidth(i);
                    restWidth-=width;
                    table.getColumn(i).setWidth(width);
                }
            }
            if(strech>=0&&strech<percent.length) {
                strechWidth = restWidth;
                if(min[strech]>strechWidth) strechWidth=min[strech];
                if(max[strech]<strechWidth) strechWidth=max[strech];
                table.getColumn(strech).setWidth(strechWidth);
            }
            layouting=false;
        }
    }

    /**
     * Calculate the width one column
     * @param i index of the column to calculate
     */
    private int calcWidth(int i) {
        if(i==strech) return strechWidth;

        int width;
        if(percent[i]>=0) width = (int)(tableWidth*percent[i]/100);
        else width = absolut[i];
        if(min[i]>width) width=min[i];
        if(max[i]<width) width=max[i];
        return width;
    }

    /**
     * Listener method called by SWT every time a column is resized via the GUI.
     * The changes made by the end user are  reflected in the layout information.
     * @param column which was resized by the GUI
     */
    public void columnResized(TableColumn column) {
        int index = table.indexOf(column);
        int width = column.getWidth();
        if(minLimit>width) width=minLimit;
        int oldWidth=calcWidth(index);
        if(width!=oldWidth) {
            if(percent[index]>=0) percent[index]=(float)width/tableWidth*100;
            else absolut[index]=width;
            if(min[index]>width) min[index]=width;
            if(max[index]<width) max[index]=width;
        } else {
            layouting=true;
            column.setWidth(width);
            layouting=false;
            return;
        }
        float percentSum=0.0f;
        int restWidth=tableWidth-width;
        for(int i=0;i<index;i++) {
            restWidth-=table.getColumn(i).getWidth();
        }
        for(int i=index+1;i<percent.length;i++) {
            if(i==strech) {
                if(width<oldWidth) restWidth-=Math.min(strechWidth+oldWidth-width,max[strech]);
                else restWidth-=Math.max(strechWidth+oldWidth-width,min[strech]);
            } else {
                if(percent[i]<0) restWidth-=table.getColumn(i).getWidth();
                else percentSum+=percent[i];
            }
        }

        float factor = ((float)restWidth/tableWidth)/(percentSum/100);
        for(int i=index+1;i<percent.length;i++) {
            if(i!=strech&&percent[i]>=0) percent[i]*=factor;
        }
        layout(true);
    }

    /**
     * Listener method called by SWT every time the table is moved. Does nothing.
     */
    public void controlMoved(ControlEvent e) {}

    /**
     * Listener method called by SWT every time the table or a column is resized.
     * If the table is resized calls {#link #layout(boolean)}.
     * If a column is resized calls {#link #columnResized}.
     */
    public void controlResized(ControlEvent e) {
        if(layouting) return; // ignore self generated events
        if(e.getSource()==table) {
            table.getDisplay().asyncExec(new Runnable() {
                public void run() {
                    if(!table.isDisposed()) {
                        layout(false);  // if it is executed inside the event method, a horizontal scrollbar apears on shrink
                    }
                }
            });
        } else if(e.getSource() instanceof TableColumn) {
            final TableColumn column = (TableColumn) e.getSource();
            table.getDisplay().asyncExec(new Runnable() {
                public void run() {
                    if(!column.isDisposed()) {
//                      if it is executed inside the event method,
//                      the horizontal scrollbar does not appear or disappear correctly
                        columnResized(column);
                    }
                }
            });
        }
    }


    /**
     * Resizes the given column to zero and scales up all columns with
     * percentage widths to fill up the gap.
     * @param index of the column to collapse
     * @since 2.1.1
     */
    public void collapseColumn(int column) {
        int oldWidth=calcWidth(column);
        int columnSum=0;
        for(int i=0;i<percent.length;i++) {
            columnSum+=calcWidth(i);
        }
        min[column]=0;
        absolut[column]=0;
        percent[column]=-1;
        if(strech==column) { strech = -1; }
        table.getColumn(column).setResizable(false);

        int restWidth;
        if(columnSum<tableWidth) restWidth=columnSum;
        else restWidth=tableWidth;
        float percentSum=0.0f;
        for(int i=0;i<percent.length;i++) {
            if(i==strech) {
                restWidth-=Math.min(strechWidth+oldWidth,max[strech]);
            } else {
                if(percent[i]<0) restWidth-=absolut[i];
                else percentSum+=percent[i];
            }
        }

        float factor = ((float)restWidth/tableWidth)/(percentSum/100);
        if(factor>1) { // only growing is allowed
            for(int i=0;i<percent.length;i++) {
                if(i!=strech&&percent[i]>=0) percent[i]*=factor;
            }
        }
        layout();
    }

}
